﻿using Eingabeverwaltung.Parser.Proof_Parser;
using Eingabeverwaltung.Metamodell;
using Antlr4.Runtime.Tree;
using Antlr4.Runtime;
using Antlr4.Runtime.Misc;
using System.Collections.Generic;

namespace Eingabeverwaltung.Parser
{
    /// <summary>
    /// Proof_Listener creates the Proof but without TransformationSteps, because 
    /// of the problems of cyclic Lemma usage or finding the topological order of uncyclic lemmata.
    /// </summary>
    public class Proof_Listener : IProofListener
    {
        // object on which will be build
        private Container container;

        // helping Listener for building
        private Listener_Helper LH;

        // Lemma parts
        private SingleProof mainProof;
        private List<Lemma> lemmata;
        private Lemma lemmaDef;

        private SingleProof SP;

        // Induction parts
        Variable indVar;
        List<InductionCase> indCases;
        List<InductionHypothesis> indHyps;

        // Transformation parts
        private Tree startTree;

        public Proof_Listener(Container container)
        {
            this.container = container;
            LH = new Listener_Helper(container.ADT.sorts, container.ADT.operations, null, container.ADT.axioms, null, null);

            mainProof = null;
            lemmata = null;
            lemmaDef = null;
            SP = null;

            indVar = null;
            indCases = null;
            indHyps = null;

            startTree = null;
        }

        // Proof controlling structures and lemmata
        public void EnterProof([NotNull] ProofParser.ProofContext context) => lemmata = new List<Lemma>();
        public void ExitProof([NotNull] ProofParser.ProofContext context) 
        {
            if (mainProof == null) throw new KontextException(new Coords(context),"No main proof found in ExitProof!");
            container.Proof = new Proof(mainProof, lemmata);
        }
        public void EnterMainProof([NotNull] ProofParser.MainProofContext context) 
        {
            // Variablen der Aufgabenstellung gelten im Hauptbeweis als erstes
            List<List<Variable>> variableDefinitionContext;
            variableDefinitionContext = new List<List<Variable>>() { container.Task.task.eq.variables };
            // Variablen des ADT gelten im Hauptbeweis als zweites, falls vorhanden
            if (container.ADT.axioms.Count != 0) variableDefinitionContext.Add(container.ADT.axioms[0].eq.variables);
            LH.setVariableContext(variableDefinitionContext);
            // feste / allQ Variablen in LH laden
            LH.fixedVars = container.Task.task.fixedVars;
            LH.allQVars = new List<Variable>(container.Task.task.eq.variables);
            foreach (Variable var in LH.fixedVars)
                LH.allQVars.Remove(var);
        }
        public void ExitMainProof([NotNull] ProofParser.MainProofContext context) => mainProof = SP;
        public void EnterLemma([NotNull] ProofParser.LemmaContext context) { }
        public void ExitLemma([NotNull] ProofParser.LemmaContext context) 
        {
            lemmata.Add(new Lemma(new Coords(context),lemmaDef.name, lemmaDef.eq, lemmaDef.fixedVars, SP)); 
            lemmaDef = null; // notwendig, damit SP weis, ob es gerade nicht in einem Lemma ist
        }
        public void EnterLemmaDef([NotNull] ProofParser.LemmaDefContext context) { }
        public void ExitLemmaDef([NotNull] ProofParser.LemmaDefContext context)
        {
            Coords coords = new Coords(context);
            string lemmaName = context.BEZ().GetText();
            Test_Lemmata_all_Unique(lemmaName,"Lemma",coords);
            lemmaDef = new Lemma(coords,lemmaName, LH.equation, LH.fixedVars,null);
            // Im Beweis des Lemmas gelten zuerst die Variablen des Lemmas (gilt bereits durch das Bauen der Gleichung)
            List<List<Variable>> variableDefinitionContext;
            variableDefinitionContext = new List<List<Variable>>() { LH.adt_vars };
            // Im Beweis des Lemmas gelten als zweites die Variablen des ADTs, falls vorhanden
            if (container.ADT.axioms.Count != 0) variableDefinitionContext.Add(container.ADT.axioms[0].eq.variables);
            LH.setVariableContext(variableDefinitionContext);
        }
        public void EnterSingleProof([NotNull] ProofParser.SingleProofContext context) => SP = null;
        public void ExitSingleProof([NotNull] ProofParser.SingleProofContext context)
        {
            if (SP == null) throw new KontextException(new Coords(context),"No SingleProof found in ExitSingleProof!");
        }
        public void EnterProof_end([NotNull] ProofParser.Proof_endContext context) { }
        public void ExitProof_end([NotNull] ProofParser.Proof_endContext context) { }

        // Transformation
        public void EnterTransformation([NotNull] ProofParser.TransformationContext context) { }
        public void ExitTransformation([NotNull] ProofParser.TransformationContext context)
        {
            SP = new Transformation(new Coords(context),startTree, new List<TransformationStep>());
        }

        public void EnterStart_tree([NotNull] ProofParser.Start_treeContext context) => LH.trees = new List<Tree>();
        public void ExitStart_tree([NotNull] ProofParser.Start_treeContext context)
        {
            if (LH.trees.Count != 1) throw new KontextException(new Coords(context),"Trees was Expected to have 1 Tree for Start_tree, but found " + LH.trees.Count + " trees! They are: " + LH.TreeListToString(LH.trees));
            startTree = LH.trees[0];
        }

        // TransformationStep is parsed by TransformationStep_Listener
        public void EnterTransformation_step([NotNull] ProofParser.Transformation_stepContext context) { }
        public void ExitTransformation_step([NotNull] ProofParser.Transformation_stepContext context) { }
        public void EnterAxiom([NotNull] ProofParser.AxiomContext context) { }
        public void ExitAxiom([NotNull] ProofParser.AxiomContext context) { }

        public void EnterDirection([NotNull] ProofParser.DirectionContext context) { }
        public void ExitDirection([NotNull] ProofParser.DirectionContext context) { }
        public void EnterTermpart([NotNull] ProofParser.TermpartContext context) { }
        public void ExitTermpart([NotNull] ProofParser.TermpartContext context) { }

        public void EnterSubst([NotNull] ProofParser.SubstContext context) { }
        public void ExitSubst([NotNull] ProofParser.SubstContext context) { }

        public void EnterSub([NotNull] ProofParser.SubContext context) { }
        public void ExitSub([NotNull] ProofParser.SubContext context) { }

        public void EnterResult_tree([NotNull] ProofParser.Result_treeContext context) { }
        public void ExitResult_tree([NotNull] ProofParser.Result_treeContext context) { }

        // Induction
        public void EnterInduction([NotNull] ProofParser.InductionContext context) 
        { indVar = null; indCases = new List<InductionCase>(); indHyps = new List<InductionHypothesis>(); }
        public void ExitInduction([NotNull] ProofParser.InductionContext context)
        {
            Coords coords = new Coords(context);
            string proofName = "";
            SingleTask lemmaIndVarTask = null;
            if (indVar == null)
            {// find indVar
                if (context.Parent.Parent.GetChild(0).GetText() == "proof")
                { indVar = container.Task.task.indVariable; proofName = "main proof"; }
                else if (context.Parent.Parent.GetChild(0).GetChild(0).GetText() == "lemma")
                {
                    proofName = ((ProofParser.LemmaDefContext)context.Parent.Parent.GetChild(0)).BEZ().GetText();
                    foreach (SingleTask lemmaTask in container.Task.lemmaTasks)
                        if (lemmaTask.name == proofName) { indVar = lemmaTask.indVariable; lemmaIndVarTask = lemmaTask; break; }
                }
                else throw new KontextException(coords,"ExitInduction with Context = " + context.GetText() + " was not able to find the sense of its life.");

            }
            if (indVar == null) throw new KontextException(coords,"Induction variable of proof "+ proofName + " is neither definied in proof nor in task!");
            Equation eq = null; List<Variable> fixedVars = null;
            if (lemmaDef != null) 
            {// this SP is in lemma lemmaDef
                eq = lemmaDef.eq;  fixedVars = lemmaDef.fixedVars;
            }else
            {// this SP is in Mainproof
                eq = container.Task.task.eq; fixedVars = container.Task.task.fixedVars;
            }
            if (lemmaIndVarTask != null)
            {   // inVar must be of context of its lemma: find substitution to get correct variable.
                Ueberpruefung.CheckHelper.getSubsAndRenamings(eq, lemmaIndVarTask.eq, out List<Substitution> subs, out List<Substitution> renamings);
                bool changed = false;
                if (subs != null) // then lemma is not proving the task (error 22), do nothing then                
                    foreach(Substitution renaming in renamings)
                        if (renaming.tree is TreeVariable tv)
                            if (tv.variable == indVar) { indVar = renaming.variable; changed = true; break; }
                if (!changed) indVar = null;
            }
            SP = new Induction(coords,eq, fixedVars,indVar,indCases, indHyps);
        }
        public void EnterInd_var([NotNull] ProofParser.Ind_varContext context)
        {
            string variableName = context.BEZ().GetText();
            Variable variable = LH.findVariableInADT_Vars(variableName);
            if (variable == null) throw new KontextException(new Coords(context),"Could not find Induction variable " + variableName + " in current context of variables.");
            indVar = variable;
        }
        public void ExitInd_var([NotNull] ProofParser.Ind_varContext context) { }
        public void EnterInd_basis([NotNull] ProofParser.Ind_basisContext context) => SP = null;
        public void ExitInd_basis([NotNull] ProofParser.Ind_basisContext context)
        {
            Coords coords = new Coords(context);
            if (SP is Transformation tr)
                indCases.Add(new InductionCase(coords,"IA", LH.equation, LH.fixedVars, tr));
            else throw new KontextException(coords,"Found Singleproof in induction basis that is not a transformation!");
        }
        public void EnterInd_step([NotNull] ProofParser.Ind_stepContext context) => SP = null;
        public void ExitInd_step([NotNull] ProofParser.Ind_stepContext context)
        {
            Coords coords = new Coords(context);
            if (SP is Transformation tr)
                indCases.Add(new InductionCase(coords, "IS", LH.equation, LH.fixedVars, tr));
            else throw new KontextException(coords,"Found Singleproof in induction step that is not a transformation!");
        }
        public void EnterInd_hyp([NotNull] ProofParser.Ind_hypContext context) { }
        public void ExitInd_hyp([NotNull] ProofParser.Ind_hypContext context)
        {
            Coords coords = new Coords(context);
            string name;
            if (context.BEZ() != null)
                name = context.BEZ().GetText();
            else name = "IH";
            foreach (InductionHypothesis ih in indHyps)
                if (ih.name == name) throw new KontextException(coords,"There are more then 1 inductionhypothesis with name "+name+"!");
            Test_Lemmata_all_Unique(name,"InductionHypothesis",coords);
            indHyps.Add(new InductionHypothesis(coords,name,LH.equation, LH.fixedVars)); 
        
        }
        public void EnterInd_case([NotNull] ProofParser.Ind_caseContext context) { }
        public void ExitInd_case([NotNull] ProofParser.Ind_caseContext context)
        {
            // Im Beweis gelten zuerst die Variablen der Gleichung (gilt bereits durch das Bauen der Gleichung)
            List<List<Variable>> variableDefinitionContext;
            variableDefinitionContext = new List<List<Variable>>() { LH.adt_vars };
            // Im Beweis gelten als zweites die Variablen des ADTs, falls vorhanden
            if (container.ADT.axioms.Count != 0) variableDefinitionContext.Add(container.ADT.axioms[0].eq.variables);
            LH.setVariableContext(variableDefinitionContext);
        }

        // Equation Rules
        public void EnterEq([NotNull] ProofParser.EqContext context) => LH.EnterEq();
        public void ExitEq([NotNull] ProofParser.EqContext context) => LH.ExitEq(new Coords(context));
        public void EnterEq_var_def([NotNull] ProofParser.Eq_var_defContext context) => LH.EnterVars();
        public void ExitEq_var_def([NotNull] ProofParser.Eq_var_defContext context) => LH.ExitEqVarDef();
        public void EnterFixed([NotNull] ProofParser.FixedContext context) => LH.EnterVars();
        public void ExitFixed([NotNull] ProofParser.FixedContext context) => LH.fixedVars = LH.adt_vars;
        public void EnterForall([NotNull] ProofParser.ForallContext context) => LH.EnterVars();
        public void ExitForall([NotNull] ProofParser.ForallContext context) => LH.allQVars = LH.adt_vars;
        public void EnterVar([NotNull] ProofParser.VarContext context) { }
        public void ExitVar([NotNull] ProofParser.VarContext context)
            => LH.ExitVar(context.BEZ().GetText(), context.sort().BEZ().GetText(),new Coords(context));
        public void EnterSort([NotNull] ProofParser.SortContext context) { }
        public void ExitSort([NotNull] ProofParser.SortContext context) { }

        // Trees
        public void EnterMytree([NotNull] ProofParser.MytreeContext context) { }
        public void ExitMytree([NotNull] ProofParser.MytreeContext context) { }

        public void EnterEq_s([NotNull] ProofParser.Eq_sContext context) { }
        public void ExitEq_s([NotNull] ProofParser.Eq_sContext context) { }
        // classic Trees
        public void EnterClassicTree([NotNull] ProofParser.ClassicTreeContext context) { }
        public void ExitClassicTree([NotNull] ProofParser.ClassicTreeContext context) { }

        public void EnterTreeLeaf([NotNull] ProofParser.TreeLeafContext context) { }
        public void ExitTreeLeaf([NotNull] ProofParser.TreeLeafContext context)
            => LH.ExitTreeLeaf(context.BEZ().GetText(),new Coords(context)); 

        public void EnterTreeNode([NotNull] ProofParser.TreeNodeContext context) { }
        public void ExitTreeNode([NotNull] ProofParser.TreeNodeContext context)
            => LH.ExitTreeNode(context.BEZ().GetText(), context.ChildCount, context.GetText(), new Coords(context));

        public void EnterFTree([NotNull] ProofParser.FTreeContext context) { }
        public void ExitFTree([NotNull] ProofParser.FTreeContext context) { }

        public void EnterFTreeRoot([NotNull] ProofParser.FTreeRootContext context) { }
        public void ExitFTreeRoot([NotNull] ProofParser.FTreeRootContext context)
        => LH.ExitFTreeRoot(new Coords(context));

        public void EnterFTreeLeft([NotNull] ProofParser.FTreeLeftContext context) { }
        public void ExitFTreeLeft([NotNull] ProofParser.FTreeLeftContext context) { }
        public void EnterFTreeRight([NotNull] ProofParser.FTreeRightContext context) { }
        public void ExitFTreeRight([NotNull] ProofParser.FTreeRightContext context) { }

        public void EnterFTreeBracket([NotNull] ProofParser.FTreeBracketContext context) { }
        public void ExitFTreeBracket([NotNull] ProofParser.FTreeBracketContext context) { }

        public void EnterFTreeBez([NotNull] ProofParser.FTreeBezContext context) { }

        public void ExitFTreeBez([NotNull] ProofParser.FTreeBezContext context)
            => LH.ExitFTreeBez(context.BEZ().GetText(), new Coords(context));

        // useless Interface requirements
        public void EnterEveryRule(ParserRuleContext ctx) { }
        public void ExitEveryRule(ParserRuleContext ctx) { }
        public void VisitErrorNode(IErrorNode node) { }
        public void VisitTerminal(ITerminalNode node) { }

        // Alle Offiziell notwendige Typtests

        private void Test_Lemmata_all_Unique(string name, string type, Coords coords)
        {
            foreach (Rule axiom in LH.adt_axioms)
                if (axiom.name == name) throw new KontextException(coords,"The " + type + " "+ name + " has same name like 1 Axiom of ADT!");
            foreach (Rule lemma in lemmata)
                if (lemma.name == name) throw new KontextException(coords,"The " + type + " "+ name + " has same name like 1 Lemma of Proof!");
            foreach (Rule lemma in container.Task.lemmata)
                if (lemma.name == name) throw new KontextException(coords,"The " + type + " "+ name + " has same name like 1 Lemma of Task!");
        }
    }
}
